import 'dart:async';
import 'dart:typed_data';

import 'package:flutter_ble_lib/flutter_ble_lib.dart';
import 'package:hiflex/hiflex_sampling.dart';
import 'package:hiflex/hiflex_raw_device.dart';

enum HiflexDeviceState {
  none,
  connected,
  disconnected,
  unknown,
}

class HiflexDeviceRawInfo {
  final String modelNumber;
  final String hardwareRevision;
  final String firmwareRevision;

  HiflexDeviceRawInfo({
    this.modelNumber = 'Unknown',
    this.hardwareRevision = 'Unknown',
    this.firmwareRevision = 'Unknown',
  });
}

class HiflexDevice {
  String get identifier => "Unknown";
  String get name => "";

  int get batteryLevel => 0;

  HiflexDeviceRawInfo getRawInfo() => HiflexDeviceRawInfo();

  Future<bool> connect() async => false;
  Future<void> disconnect() async {}

  Stream<HiflexDeviceState> get connectionState => Stream.empty();

  Future<SamplingData> get sampling async => samplingStream().first;
  Stream<SamplingData> samplingStream() => Stream.empty();
}

class HiflexDeviceBT implements HiflexDevice {
  static final String btSerialService = "6e400000-b5a3-f393-e0a9-e50e24dcca9e";
  static final String btTxCharacteristic =
      "6e400002-b5a3-f393-e0a9-e50e24dcca9e";
  static final String btRxCharacteristic =
      "6e400003-b5a3-f393-e0a9-e50e24dcca9e";

  static final String btFtsameService = "0000180a-0000-1000-8000-00805f9b34fb";
  static final String btModelNumber = "00002a24-0000-1000-8000-00805f9b34fb";
  static final String btHardwareRevision =
      "00002a27-0000-1000-8000-00805f9b34fb";
  static final String btFirmwareRevision =
      "00002a26-0000-1000-8000-00805f9b34fb";

  HiflexRawDevice _rawDevice;
  HiflexSampling _sampling = new HiflexSampling();

  Service? _serialService;
  Characteristic? _serialTx;
  Characteristic? _serialRx;

  final StreamController<HiflexDeviceState> _connectionState =
      StreamController.broadcast();

  final String modelNumber;
  final String hardwareRevision;
  final String firmwareRevision;

  StreamController<SamplingData> _sc = StreamController.broadcast();

  int _batteryLevel = 0;

  HiflexDeviceBT(this._rawDevice, this.modelNumber, this.hardwareRevision,
      this.firmwareRevision);

  String get identifier => _rawDevice.peripheral?.identifier ?? "Unknown";
  String get name => _rawDevice.name;
  Peripheral? get _peripheral => _rawDevice.peripheral ?? null;

  HiflexDeviceRawInfo getRawInfo() {
    return HiflexDeviceRawInfo(
      modelNumber: modelNumber,
      hardwareRevision: hardwareRevision,
      firmwareRevision: firmwareRevision,
    );
  }

  Future<bool> connect() async {
    Future<Service> findService(String targetUUID) async {
      var services = await _peripheral?.services();
      if (services != null) {
        for (var service in services) {
          print("Service: {${service.uuid}}");
          if (service.uuid.compareTo(targetUUID) == 0) {
            return service;
          }
        }
      }
      return Stream<Service>.empty().first;
    }

    Future<Characteristic> findCharacteristic(String targetUUID) async {
      var characteristics = await _serialService?.characteristics();
      if (characteristics != null) {
        for (var characteristic in characteristics) {
          if (characteristic.uuid == targetUUID) {
            return characteristic;
          }
        }
      }
      return Stream<Characteristic>.empty().first;
    }

    if (await _peripheral?.isConnected() ?? false) {
      _serialService = await findService(btSerialService);
      if (_serialService != null) {
        _serialTx = await findCharacteristic(btTxCharacteristic);
        _serialRx = await findCharacteristic(btRxCharacteristic);

        _serialRx?.monitor()?.listen((samplingRawData) {
          _sc.add(_sampling.load(samplingRawData));
          _batteryLevel = _sampling.batteryLevel;
        });

        _peripheral?.observeConnectionState()?.listen((event) {
          print('$name ${_peripheral?.identifier ?? "Null"} state: $event');
          switch (event) {
            case PeripheralConnectionState.connected:
              _connectionState.add(HiflexDeviceState.connected);
              break;
            case PeripheralConnectionState.disconnected:
              _connectionState.add(HiflexDeviceState.disconnected);
              break;
            default:
              break;
          }
        });

        return true;
      }
    }
    return false;
  }

  int get batteryLevel => _batteryLevel;

  Future<void> disconnect() async =>
      _peripheral?.disconnectOrCancelConnection();

  Stream<HiflexDeviceState> get connectionState {
    return _connectionState.stream;
  }

  Future<SamplingData> get sampling {
    return _sc.stream.first;
  }

  Stream<SamplingData> samplingStream() => _sc.stream;

  sendPacket(Uint8List value) {
    _serialTx?.write(value, false);
  }
}
